// $Id: ID.cc,v 1.36 2004/03/01 12:23:25 vern Exp $
//
// Copyright (c) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include "ID.h"
#include "Expr.h"
#include "Dict.h"
#include "EventRegistry.h"
#include "Func.h"
#include "Scope.h"
#include "File.h"
#include "Serializer.h"
#include "RemoteSerializer.h"
#include "Scope.h"

ID::ID(const char* arg_name, IDScope arg_scope, bool arg_is_export)
	{
	name = copy_string(arg_name);
	scope = arg_scope;
	is_export = arg_is_export;
	type = 0;
	val = 0;
	attrs = 0;
	is_const = 0;
	is_enum_const = 0;
	is_type = 0;
	offset = 0;

	infer_return_type = false;

	SetLocationInfo(&start_location, &end_location);
	}

ID::~ID()
	{
	delete [] name;
	Unref(type);
	Unref(val);
	Unref(attrs);
	}

string ID::ModuleName() const
	{
	return extract_module_name(name);
	}

void ID::SetVal(Val* v, Opcode op)
	{
	// Send update for synchronized values.
	if ( attrs && attrs->FindAttr(ATTR_SYNCHRONIZED) )
		{
		if ( v->IsMutableVal() )
			v->AsMutableVal()->AddProperties(MutableVal::SYNCHRONIZED);
		remote_serializer->SendAccess(StateAccess(op, this, v, val));
		}

	Unref(val);
	val = v;

	if ( type && val &&
	     type->Tag() == TYPE_FUNC && type->AsFuncType()->IsEvent() )
		{
		EventHandler* handler = event_registry->Lookup(name);
		if ( ! handler )
			{
			handler = new EventHandler(name);
			handler->SetLocalHandler(val->AsFunc());
			event_registry->Register(handler);
			}
		else
			{
			// Otherwise, internally defined events cannot
			// have local handler.
			handler->SetLocalHandler(val->AsFunc());
			}
		}
	}

void ID::SetVal(Val* v, init_class c)
	{
	if ( c == INIT_NONE || c == INIT_FULL )
		{
		SetVal(v);
		return;
		}

	if ( type->Tag() != TYPE_TABLE &&
	     (type->Tag() != TYPE_PATTERN || c == INIT_REMOVE) )
		{
		if ( c == INIT_EXTRA )
			Error("+= initializer only applies to tables, sets and patterns", v);
		else
			Error("-= initializer only applies to tables and sets", v);
		}

	else
		{
		if ( c == INIT_EXTRA )
			{
			if ( ! val )
				SetVal(v);
			else
				v->AddTo(val, 0);
			}
		else
			{
			if ( val )
				v->RemoveFrom(val);
			}
		}

	Unref(v);
	}

void ID::SetVal(Expr* ev, init_class c)
	{
	Attr* a = attrs->FindAttr(c == INIT_EXTRA ?
					ATTR_ADD_FUNC : ATTR_DEL_FUNC);

	if ( ! a )
		Internal("no add/delete function in ID::SetVal");

	EvalFunc(a->AttrExpr(), ev);
	}

void ID::SetAttrs(Attributes* a)
	{
	Unref(attrs);
	attrs = a;

	if ( val && val->Type()->Tag() == TYPE_TABLE )
		val->AsTableVal()->SetAttrs(attrs);
	
	if ( val && val->Type()->Tag() == TYPE_FILE )
		val->AsFile()->SetAttrs(attrs);
	}

void ID::AddAttrs(Attributes* a)
	{
	if ( ! a )
		return;

	if ( a->FindAttr(ATTR_SYNCHRONIZED)	)
		{
		remote_serializer->RegisterID(this);
		if ( val && val->IsMutableVal() )
			val->AsMutableVal()->AddProperties(MutableVal::SYNCHRONIZED);
		}

	if ( a->FindAttr(ATTR_PERSISTENT)   )
		{
		persistence_serializer->Register(this);
		if ( val && val->IsMutableVal() )
			val->AsMutableVal()->AddProperties(MutableVal::PERSISTENT);
		}

	if ( attrs )
		attrs->AddAttrs(a);
	else
		attrs = a;

	if ( val && val->Type()->Tag() == TYPE_TABLE )
		val->AsTableVal()->SetAttrs(attrs);

	if ( val && val->Type()->Tag() == TYPE_FILE )
		val->AsFile()->SetAttrs(attrs);
	}

void ID::EvalFunc(Expr* ef, Expr* ev)
	{
	Expr* arg1 = new ConstExpr(val->Ref());
	ListExpr* args = new ListExpr();
	args->Append(arg1);
	args->Append(ev->Ref());

	CallExpr* ce = new CallExpr(ef->Ref(), args);

	SetVal(ce->Eval(0));
	Unref(ce);
	}

bool ID::Serialize(Serializer* s, bool bind_to_current,
			const SerialInfo& info) const
	{
	if ( info.cont )
		return val->Serialize(s, true, info);

	assert(type);

	s->WriteOpenTag("ID");

	if ( ! (s->Write(name) &&
		s->Write((char)scope) &&
		s->Write(is_export) &&
		s->Write(is_const) &&
		s->Write(is_enum_const) &&
		s->Write(is_type) &&
		s->Write(offset) &&
		type->Serialize(s)) )
		return false;

	SERIALIZE_OPTIONAL(s, attrs)

	if ( ! bind_to_current )
		{
		if ( val )
			{
			if ( ! (s->Write(true) && val->Serialize(s, true, info)) )
				return false;
			}

		else if ( ! (s->Write(false)) )
			return false;
		}

	else if ( ! s->Write(val!=0) )
		return false;

	s->WriteCloseTag("ID");
	return true;
	}

ID* ID::Unserialize(Serializer* s, bool bind_to_current)
	{
	bool installed_tmp = false;
	ID* id = new ID();
	ID* current = 0;
	bool had_val = false;

	char id_scope;

	if ( ! (s->Read(&id->name, 0) &&
		s->Read(&id_scope) &&
		s->Read(&id->is_export) &&
		s->Read(&id->is_const) &&
		s->Read(&id->is_enum_const) &&
		s->Read(&id->is_type) &&
		s->Read(&id->offset)
	       ) )
		goto error;

	id->scope = IDScope(id_scope);

	s->SetErrorDescr(fmt("unserializing ID %s", id->name));

	id->type = BroType::Unserialize(s);
	if ( ! id->type )
		goto error;

	UNSERIALIZE_OPTIONAL_STATIC(s, id->attrs, Attributes::Unserialize(s), id);

	current = global_scope()->Lookup(id->name);

	if ( ! bind_to_current )
		{
		// If it's a global function not currently known,
		// we temporarily install it in global scope.
		// This is neccessary for recursive functions.
		if ( ! current && id->IsGlobal() &&
		     id->Type()->Tag() == TYPE_FUNC )
			{
			installed_tmp = true;
			global_scope()->Insert(id->Name(), id);
			}

		bool have_val;
		if ( ! s->Read(&have_val) )
			goto error;

		if ( have_val )
			{
			id->val = Val::Unserialize(s, 0, TYPE_ANY);
			if ( ! id->val )
				goto error;
			}
		}
	else
		{
		id->val = 0;
		if ( ! s->Read(&had_val) )
			return false;
		}

	if ( id->IsGlobal() && bind_to_current && ! s->IsPrintOnly() )
		{
		// Perform type checking for already existing global IDs.
		if ( ! current )
			{
			// Reference to non-existent global. This is ok if
			// it didn't exist during serialization either.
			if ( ! had_val )
				goto ok;

			s->Error("undefined");
			goto error;
			}

		if ( ! same_type(current->type, id->type) )
			{
			s->Error("type mismatch for id");
			goto error;
			}

		// We're not interested in the stored ID's value but
		// but reference the current global.
		delete id;
		Ref(current);
		id = current;
		goto ok;
		}

	if ( id->val && id->val->Type()->Tag() == TYPE_FUNC &&
	     id->val->AsFunc() )
		// We know that all our functions are in fact debug functions...
		((DebugFunc*)id->val->AsFunc())->SetID(id);

ok:
	if ( installed_tmp && ! global_scope()->Remove(id->name) )
		internal_error("tmp id missing");

	return id;

error:
	if ( installed_tmp && ! global_scope()->Remove(id->name) )
		internal_error("tmp id missing");

	delete id;
	return 0;
	}


void ID::Error(const char* msg, const BroObj* o2)
	{
	BroObj::Error(msg, o2, 1);
	SetType(error_type());
	}

void ID::Describe(ODesc* d) const
	{
	d->Add(name);
	}
