// $Id: PktSrc.h,v 1.13 2004/01/20 22:00:18 vern Exp $
//
// Copyright (c) 1999, 2000, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#ifndef pktsrc_h
#define pktsrc_h

#include "Dict.h"
#include "Expr.h"

extern "C" {
#include <pcap.h>
}

declare(PDict,bpf_program);

// Whether a PktSrc object is used by the normal filter structure or the
// secondary-path structure.
typedef enum {
	TYPE_FILTER_NORMAL,  // the normal filter
	TYPE_FILTER_SECONDARY,  // the secondary-path filter
} PktSrc_Filter_Type;


// {filter,event} tuples conforming the secondary path.
class SecondaryEvent {
public:
	SecondaryEvent(const char* arg_filter, Func* arg_event)
		{
		filter = arg_filter;
		event = arg_event;
		}

	const char* Filter()	{ return filter; }
	Func* Event()		{ return event; }

private:
	const char* filter;
	Func* event;
};

declare(PList,SecondaryEvent);
typedef PList(SecondaryEvent) secondary_event_list;



class SecondaryPath {
public:
	SecondaryPath();
	~SecondaryPath();

	secondary_event_list& EventTable()	{ return event_list; }
	const char* Filter()			{ return filter; }

private:
	secondary_event_list event_list;
	// OR'ed union of all SecondaryEvent filters
	char* filter;
};

// Main secondary-path object.
extern SecondaryPath* secondary_path;


// {program, {filter,event}} tuple table.
class SecondaryProgram {
public:
	SecondaryProgram(struct bpf_program* arg_program, SecondaryEvent* arg_event)
		{
		program = arg_program;
		event = arg_event;
		}

	~SecondaryProgram();

	struct bpf_program* Program()	{ return program; }
	SecondaryEvent* Event()		{ return event; }

private:
	// Associated program.
	struct bpf_program* program;

	// Event that is run in case the program is matched.
	SecondaryEvent* event;
};

declare(PList,SecondaryProgram);
typedef PList(SecondaryProgram) secondary_program_list;



class PktSrc {
public:
	~PktSrc();

	int NotOpen() const		{ return pd == 0; }
	const char* ErrorMsg() const	{ return errbuf; }
	void ClearErrorMsg()		{ *errbuf ='\0'; }

	pcap_dumper_t* OpenDumper(const char* writefile);

	// Returns nil if there are no more packets.  Will continue to
	// return the same packet until ConsumePacket() is called; this
	// allows for "peeking" at the next packet but deferring processing
	// until later if needed.
	const u_char* NextPacket(struct pcap_pkthdr*& hdr);

	// Writes the last read packet into file.
	bool DumpCurrentPacket(const char* file);

	// Same, but just for the timestamp of the next packet.
	// Returns < 0 if no more packets.
	double NextTimestamp();

	int HdrSize() const		{ return hdr_size; }
	int DataLink() const		{ return datalink; }

	void ConsumePacket()	{ data = 0; }

	int IsLive() const		{ return interface != 0; }

	pcap_t* PcapHandle() const	{ return pd; }

	const char* ReadFile() const	{ return readfile; }
	const char* Interface() const	{ return interface; }
	const PktSrc_Filter_Type FilterType() const	{ return filter_type; }
	void AddSecondaryTablePrograms();
	const secondary_program_list& ProgramTable() const
		{ return program_list; }


	// Precompiles a filter and associates the given index with it.
	// Returns true on success, 0 if a problem occurred.
	int PrecompileFilter(int index, const char* filter);

	// Activates the filter with the given index.
	// Returns true on success, 0 if a problem occurred.
	int SetFilter(int index);

protected:
	PktSrc();

	void SetHdrSize();

	void Close();

	struct pcap_pkthdr hdr;
	const u_char* data;	// contents of current packet
	const u_char* last_data;	// same, but unaffected by consuming
	int hdr_size;
	int datalink;

	char* interface;	// nil if not reading from an interface
	char* readfile;		// nil if not reading from a file

	pcap_t* pd;
	uint32 netmask;
	char errbuf[PCAP_ERRBUF_SIZE + 256];

	PDict(bpf_program) filters; // precompiled filters

	PktSrc_Filter_Type filter_type; // normal path or secondary path
	secondary_program_list program_list;

	const char* current_dump_file; // "cache" for DumpCurrentPacket()
	pcap_dumper_t* current_dumper;
};

class PktInterfaceSrc : public PktSrc {
public:
	PktInterfaceSrc(const char* interface, const char* filter,
			PktSrc_Filter_Type ft=TYPE_FILTER_NORMAL);
};

class PktFileSrc : public PktSrc {
public:
	PktFileSrc(const char* readfile, const char* filter,
			PktSrc_Filter_Type ft=TYPE_FILTER_NORMAL);
};

#ifdef DONT_HAVE_LIBPCAP_PCAP_FREECODE
extern "C" {
	int pcap_freecode(pcap_t*, struct bpf_program*);
	int pcap_compile_nopcap(int, int, struct bpf_program*,
			char*, int, bpf_u_int32);
}
#endif

#endif
