// $Id: RemoteSerializer.h,v 1.5 2004/03/01 12:23:25 vern Exp $
//
// Communication between two Bro's.

#include "Dict.h"
#include "List.h"
#include "Serializer.h"
#include "Event.h"


// #define DEBUG_COMMUNICATION

// All IP arguments are in host byte-order.
// FIXME: Change this to network byte order

// This class handles the communication done in Bro's main loop.
class RemoteSerializer : public Serializer, public EventSource {
public:
	RemoteSerializer();
	virtual ~RemoteSerializer();

	// Connect to host.
	bool Connect(addr_type ip, uint16 port, double retry, bool use_ssl);

	// Request all events matching pattern from remote side.
	bool Request(addr_type ip, uint16 port, RE_Matcher* pattern);

	// Sets flag whether we're accepting state from this peer
	// (default: yes).
	bool SetAcceptState(addr_type ip, uint16 port, bool accept);

	// Start to listen.
	bool Listen(addr_type ip, uint16 port, bool expect_ssl);

	// Stop it.
	bool StopListening();

	// Send the event/function call to the given address
	// (if already connected).
	bool SendCall(addr_type ip, uint16 port, const char* name, val_list* vl);

	// Sends ID.
	bool SendID(addr_type ip, uint16 port, const ID& id);

	// Broadcasts the access.
	bool SendAccess(const StateAccess& access);

	// Sends the internal connection state.
	bool SendConnection(addr_type ip, uint16 port, const Connection& c);

	// Check for incoming events and queue them.
	bool Poll(bool may_block);

	// Returns the corresponding record (already ref'ed).
	RecordVal* GetEventSourceVal(Event::SourceID id);

	// Send our capture filter to the remote peer.
	bool SendCaptureFilter(addr_type ip, uint16 port, const char* filter);

	// Registers the ID to be &synchronized.
	void RegisterID(ID* id)	{ sync_ids.append(id); }

	// Sends this newly synchronized ID including its Val to all peers.
	bool SendSynchronizedID(ID* id);

	// All received events are put into a local buffer first, accompanied
	// by the time of reception.

	// Overidden from EventSource:

	// Dispatch the oldest event and remove it from the buffer.
	virtual void DispatchNextEvent();

	// Return the time of the oldest event (0 if buffer is empty).
	virtual double NextTimestamp();

	// Return true if we're either listening or there is least one
	// open connection.
	virtual bool IsActive();

	// Log some statistics.
	void LogStats();

protected:
	void Log(const char* msg);
	// Send ID. This is not for general use, though, but just for
	// some IDs with special interpretation.
	bool SendID(addr_type ip, uint16 port, const char* name,
			bool force_sync_val = false);

	virtual void ReportError(const char* msg);
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args);
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args);
	virtual void GotID(ID* id, Val* val);
	virtual void GotStateAccess(StateAccess* s);
	virtual void GotTimer(Timer* t);
	virtual void GotConnection(Connection* c);
	virtual void UnknownID(ID* id);

private:
	declare(PList, EventHandler);
	typedef PList(EventHandler) handler_list;

	struct Peer {
		Event::SourceID id; // Unique ID (non-zero) per peer.
		addr_type ip;
		uint16 port;
		handler_list handlers;
		RecordVal* val;		// Record of type event_source.
		SerializationCache cache_in; 	// One cache for each direction.
		SerializationCache cache_out;
		bool connected;
		bool orig;	// True if we connected to the peer.
		bool accept_state; // True if we accept state from the peer.
	};

	void Init();

	void Fork();

	bool ProcessSerialization();
	bool ProcessRequestMsg();
	bool ProcessErrorMsg();
	bool ProcessStatsMsg();
	bool ProcessCaptureFilterMsg();

	bool SendAllSynchronized(Peer *p);
	bool SendAccess(Peer* p, const StateAccess& access);

	Peer* AddPeer(addr_type ip, uint16 port, bool connected);
	Event::SourceID RemovePeer(addr_type ip, uint16 port);
	Peer* LookupPeer(addr_type ip, uint16 port, bool only_if_connected);
	void UnregisterHandlers(Peer* peer);

	void RaiseEvent(EventHandlerPtr event, addr_type ip,
			uint16 port, const char* arg = 0);

	bool initialized;
	bool listening;
	bool ignore_accesses;
	int error_fd;
	double poll_time;	// Time when we last polled for input.

	addr_type current_ip;	// Peer from which we received the last message.
	addr_type current_port;
	char state;	// Current communication state; see MSG_* in RemoteSerializer.cc.
	Event::SourceID id_counter;	// Keeps track of assigned IDs.

	declare(PDict, Peer);
	typedef PDict(Peer) peer_map;
	declare(PList, Peer);
	typedef PList(Peer) peer_list;

	peer_map peers_byaddr;
	peer_list peers;

	struct BufferedEvent {
		time_t time;
		Event::SourceID src;
		EventHandlerPtr handler;
		val_list* args;
	};

	declare(PList, BufferedEvent);
	typedef PList(BufferedEvent) EventQueue;
	EventQueue events;

	id_list sync_ids;

	// Keep some stats
	unsigned long events_received;
	unsigned long events_sent;
	};


// This class handles the communication done in the forked child.
class SocketComm {
public:
	SocketComm(int errorfd);
	~SocketComm();

	void SetParentIO(ChunkedIO* arg_io)	{ io = arg_io; }

	void Run();	// does not return

	// Log some statistics (via pipe to parent).
	void LogStats();

protected:
	struct Peer {
		addr_type ip;
		uint16 port;	// host byte-order
		ChunkedIO* io;
		char state;
		bool connected;
		bool ssl;
		// If we get disconnected, reconnect after this many seconds.
		int retry;
		// Time of next connection attempt (0 if none).
		time_t next_try;
	};

	void Listen(addr_type ip, uint16 port, bool expect_ssl);
	void AcceptConnection(int listen_fd);
	Peer* Connect(addr_type ip, uint16 port, unsigned int retry, bool use_ssl);
	void RequestEvents(addr_type ip, RE_Matcher* pattern);
	void CloseConnection(Peer* peer, bool reconnect);

	void ProcessRemoteMessage(Peer* p);
	void ProcessParentMessage();

	void Log(const char* msg);
	bool SendError(const char* msg, addr_type ip, uint16 port);

	// If p is not nil, the peer caused the error and the connection
	// to it will be closed. If kill is true, this is a fatal error,
	// and the process will kill itself.
	void Error(const char* msg, Peer* p, bool kill);

	// Kill the current process.
	void Kill();

	// Peers we are communicating with:
	declare(PDict, Peer);
	typedef PDict(Peer) peer_map;
	declare(PList, Peer);
	typedef PList(Peer) peer_list;

	peer_list peers;
	peer_map peers_byaddr;

	ChunkedIO* io;	// I/O to parent
	char parent_state;
	int error_fd;
	Peer* peer;
	addr_type ip;	// ip/port contained in last msg from parrent.
	uint16 port;
	int listen_fd_clear;
	int listen_fd_ssl;

	// If the port we're trying to bind to is already in use, we will retry
	// it regularly.
	addr_type listen_if;
	uint16 listen_port;
	bool listen_ssl;
	time_t listen_next_try;
};

extern RemoteSerializer* remote_serializer;

#ifdef DEBUG_COMMUNICATION
# define DEBUG_COMMPARENT(s) fputs("PARENT: |" s "|\n", stderr);
# define DEBUG_COMMCHILD(s) fputs("CHILD: |" s "|\n", stderr);
#else
# define DEBUG_COMMPARENT(s)
# define DEBUG_COMMCHILD(s)
#endif
