// $Id: SerialObj.h,v 1.2 2004/03/15 20:13:57 vern Exp $
//
// Infrastructure for serializable objects.
//
// How to make objects of class Foo serializable:
//
//    1. Derive Foo (directly or indirectly) from SerialObj.
//    2. Add a SER_FOO constant to SerialTypes below.
//    3. Add DECLARE_SERIAL(Foo) into class definition.
//    4. Add a (preferably protected) default ctor if it doesn't already exist.
//    5. For non-abstract classes, add IMPLEMENT_SERIAL(Foo, SER_FOO) to *.cc
//    6. Add two methods like this to *.cc (keep names of arguments!)
//
//       bool Foo::DoSerialize(Serializer* s, SerialInfo* i, SerialType t) const
//           {
//           DO_SERIALIZE(SER_FOO, ParentClassOfFoo);
//           <... serialize class members via methods in Serializer ...>
//           return true if everything ok;
//	     }
//
//       bool Foo::DoUnserialize(Serializer* s)
//           {
//           DO_UNSERIALIZE(ParentClassOfFoo);
//           <... unserialize class members via methods in Serializer ...>
//           return true if everything ok;
//           }
//
//   (7. If no parent class of Foo already contains Serialize()/Unserialize()
//       methods, these need to be added somewhere too. But most of the various
//       parts of the class hierarchy already have them.)


#ifndef SERIALOBJ_H
#define SERIALOBJ_H

#include <map>
#include <util.h>

// Each serializable class gets a type.
//
// The type enables a form of poor man's type-checking:
//     Bit 0-7:  Number (unique relative to main parent (see below)).
//     Bit 8-12: Main parent class (SER_IS_*)
//     Bit 13:   1 if derived from BroObj.

typedef uint16 SerialType;

static const SerialType SER_IS_BRO_OBJ = 0x1000;

static const SerialType SER_IS_CONNECTION = 0x0100 | SER_IS_BRO_OBJ;
static const SerialType SER_IS_TIMER = 0x0200;
static const SerialType SER_IS_TCP_ENDPOINT = 0x0300;
static const SerialType SER_IS_TCP_ANALYZER = 0x0400 | SER_IS_BRO_OBJ;
static const SerialType SER_IS_TCP_ENDPOINT_ANALYZER = 0x0500 | SER_IS_BRO_OBJ;
static const SerialType SER_IS_TCP_CONTENTS = 0x0600;
static const SerialType SER_IS_REASSEMBLER = 0x0700;

// These are the externally visible types.
const SerialType SER_NONE = 0;

const SerialType SER_BRO_OBJ = 1 | SER_IS_BRO_OBJ;

const SerialType SER_CONNECTION = 1 | SER_IS_CONNECTION;
const SerialType SER_ICMP_CONNECTION = 2 | SER_IS_CONNECTION;
const SerialType SER_ICMP_ECHO = 3 | SER_IS_CONNECTION;
const SerialType SER_ICMP_UNREACH = 4 | SER_IS_CONNECTION;
const SerialType SER_TCP_CONNECTION = 5 | SER_IS_CONNECTION;
const SerialType SER_TCP_CONNECTION_CONTENTS = 6 | SER_IS_CONNECTION;
const SerialType SER_FTP_CONN = 7 | SER_IS_CONNECTION;

const SerialType SER_UDP_CONNECTION = 8 | SER_IS_CONNECTION;

const SerialType SER_TIMER = 1 | SER_IS_TIMER;
const SerialType SER_CONNECTION_TIMER = 2 | SER_IS_TIMER;

const SerialType SER_TCP_ENDPOINT = 1 | SER_IS_TCP_ENDPOINT;

const SerialType SER_TCP_ANALYZER = 1 | SER_IS_TCP_ANALYZER;

const SerialType SER_TCP_ENDPOINT_ANALYZER = 1 | SER_IS_TCP_ENDPOINT_ANALYZER;

const SerialType SER_TCP_CONTENTS = 1 | SER_IS_TCP_CONTENTS;
const SerialType SER_TCP_CONTENT_LINE = 2 | SER_IS_TCP_CONTENTS;
const SerialType SER_TCP_NVT = 3 | SER_IS_TCP_CONTENTS;

const SerialType SER_REASSEMBLER = 1 | SER_IS_REASSEMBLER;
const SerialType SER_TCP_REASSEMBLER = 2 | SER_IS_REASSEMBLER;

class Serializer;
class SerialInfo;

// Per-process unique ID.
class TransientID {
public:
	TransientID()	{ id = ++counter; }

	typedef uint32 ID;
	ID Value() const	{ return id; }

private:
	ID id;
	static ID counter;
};


// Abstract base class for serializable objects.
class SerialObj {
public:
	virtual ~SerialObj()	{ }

	// Instantiate an object of the given type. Return nil
	// if unknown.
	static SerialObj* Instantiate(SerialType type);

	// Associate a "factory" function with the given type.
	// A factory is a class or function that creates instances
	// of a certain type.

	typedef SerialObj* (*FactoryFunc)();
	static void Register(SerialType type, FactoryFunc f);

	static bool IsBroObj(SerialType type)
		{ return type & SER_IS_BRO_OBJ; }

	// Serializes this object completely.
	// If cache is false, no object caching is used. Then we can use
	// DECLARE_NON_CACHEABLE_SERIAL (instead of DECLARE_SERIAL) which
	// avoids storing a per-object id.
	bool Serialize(Serializer* s, SerialInfo* i, bool cache = true) const;

	// Unserializes next object.
	static SerialObj* Unserialize(Serializer* s, SerialType type,
					bool cache = true);

protected:
	SerialObj()	{ }

	virtual bool DoSerialize(Serializer* s, SerialInfo* info,
					SerialType type) const;
	virtual bool DoUnserialize(Serializer *s);

	virtual const TransientID* GetTID() const	{ return 0; }

	typedef std::map<SerialType, FactoryFunc> FactoryMap;
	static FactoryMap* factories;
};

// A class that registers a factory function upon instantiation.
class SerialTypeRegistrator {
public:
	SerialTypeRegistrator(SerialType type, SerialObj::FactoryFunc func)
		{
		SerialObj::Register(type, func);
		}
};


// Macro helpers.

#define DECLARE_SERIAL(classname) \
	static classname* Instantiate(); \
	static SerialTypeRegistrator register_type; \
	virtual bool DoSerialize(Serializer*, SerialInfo* = 0, \
					SerialType = SER_NONE) const; \
	virtual bool DoUnserialize(Serializer*); \
	virtual const TransientID*  GetTID() const	{ return &tid; } \
	TransientID tid;

#define DECLARE_NON_CACHEABLE_SERIAL(classname) \
	static classname* Instantiate(); \
	static SerialTypeRegistrator register_type; \
	virtual bool DoSerialize(Serializer*, SerialInfo* = 0, \
					SerialType = SER_NONE) const; \
	virtual bool DoUnserialize(Serializer*); \


// Only needed (and usable) for non-abstract classes.
#define IMPLEMENT_SERIAL(classname, classtype) \
	SerialTypeRegistrator classname::register_type(classtype, \
				FactoryFunc(&classname::Instantiate)); \
	classname* classname::Instantiate()	{ return new classname(); } \

#define DO_SERIALIZE(classtype, super) \
	{ \
	if ( ! super::DoSerialize(s, i, t == SER_NONE ? classtype : t) ) \
		return false; \
	} 

#define DO_UNSERIALIZE(super) \
	{ \
	if ( ! super::DoUnserialize(s) ) \
		return false; \
	}

// Some helpers for pointers which may be nil.
#define SERIALIZE_OPTIONAL(s, ptr) \
	{	\
	if ( ptr )	\
		{	\
		if ( ! (s->Write(true) && ptr->Serialize(s)) )	\
			return false;	\
		}	\
	\
	else if ( ! s->Write(false) )	\
		return false;	\
	}

#define SERIALIZE_OPTIONAL_STR(s, str) \
	{	\
	if ( str )	\
		{	\
		if ( ! (s->Write(true) && s->Write(str)) )	\
			return false;	\
		}	\
	\
	else if ( ! s->Write(false) )	\
		return false;	\
	}

#define SERIALIZE_OPTIONAL_VAL(s, ptr, t) \
	{	\
	if ( ptr )	\
		{	\
		SerialInfo no_incremental;	\
		if ( ! (s->Write(true) && \
			ptr->Serialize(s, t, no_incremental)) )	\
			return false;	\
		}	\
	\
	else if ( ! s->Write(false) )	\
		return false;	\
	}

#define UNSERIALIZE_OPTIONAL(s, dst, unserialize)	\
	{	\
	bool has_it;	\
	if ( ! s->Read(&has_it) )	\
		return false;	\
	\
	if ( has_it )	\
		{	\
		dst = unserialize;	\
		if ( ! dst )	\
			return false;	\
		}	\
	\
	else	\
		dst = 0;	\
	}

#define UNSERIALIZE_OPTIONAL_STR(s, dst)	\
	{	\
	bool has_it;	\
	if ( ! s->Read(&has_it) )	\
		return false;	\
	\
	if ( has_it )	\
		{	\
		s->Read(&dst, 0);	\
		if ( ! dst )	\
			return false;	\
		}	\
	\
	else	\
		dst = 0;	\
	}

#define UNSERIALIZE_OPTIONAL_STATIC(s, dst, unserialize, del)	\
	{	\
	bool has_it;	\
	if ( ! s->Read(&has_it) )	\
		{	\
		delete del;	\
		return 0;	\
		}	\
	\
	if ( has_it )	\
		{	\
		dst = unserialize;	\
		if ( ! dst )	\
			{	\
			delete del;	\
			return 0;	\
			}	\
		}	\
	\
	else	\
		dst = 0;	\
	}

#define SERIALIZE_BIT(s, bit) \
	s->Write(bool(bit))

#define UNSERIALIZE_BIT(s, bit) \
	{ \
	bool tmp; \
	if ( ! s->Read(&tmp) ) \
		return false; \
	bit = (unsigned int) tmp; \
	}

#endif
