// $Id: Serializer.cc,v 1.14 2004/03/10 06:34:58 vern Exp $

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <libgen.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "Serializer.h"
#include "Scope.h"
#include "Stmt.h"
#include "Logger.h"
#include "Func.h"
#include "Event.h"
#include "EventRegistry.h"
#include "SerializationFormat.h"
#include "NetVar.h"
#include "Conn.h"
#include "Timer.h"
#include "RemoteSerializer.h"

Serializer::Serializer(SerializationFormat* arg_format)
	{
	if ( arg_format )
		format = arg_format;
	else
		format = new BinarySerializationFormat();

	io = 0;
	verbose = false;
	error_descr = 0;
	current_cache = 0;
	print_only = false;
	}

Serializer::~Serializer()
	{
	delete format;
	delete [] error_descr;
	}

bool Serializer::Read(string* s)
	{
	char* cstr;
	int len;
	if ( format->Read(&cstr, &len) )
		{
		s->assign(cstr, len);
		delete [] cstr;
		return true;
		}
	else
		return false;
	}

bool Serializer::StartSerialization(const char *descr, char tag)
	{
	format->StartWrite();
	assert(current_cache);
	SetErrorDescr(fmt("serializing %s", descr));
	if ( ! Write(tag) )
		{
		Error(io->Error());
		return false;
		}
	return true;
	}

bool Serializer::EndSerialization()
	{
	ChunkedIO::Chunk* chunk = new ChunkedIO::Chunk;
	chunk->len = format->EndWrite(&chunk->data);
	if ( ! io->Write(chunk) )
		{
		Error(io->Error());
		return false;
		}
	return true;
	}

bool Serializer::Serialize(const ID& id, const SerialInfo& info)
	{
	if ( ! (id.IsGlobal() || id.IsEnumConst()) )
		{
		Error("non-global identifiers cannot be serialized");
		return false;
		}

	if ( ! info.cont && ! StartSerialization("ID", 'i') )
			return false;

	if ( info.incr )
		*info.incr = false;

	if ( ! id.Serialize(this, false, info) )
		{
		Error("failed");
		return false;
		}

	if ( (! info.incr) || (! *info.incr) )
		{
		WriteSeparator();
		return EndSerialization();
		}

	return true;
	}

bool Serializer::Serialize(const char* func, val_list* args)
	{
	if ( ! StartSerialization("call", 'e') )
		return false;

	WriteOpenTag("call");
	int a = args->length();
	Write(func, "name");
	Write(network_time, "time");
	Write(a);

	SerialInfo info;
	loop_over_list(*args, i) (*args)[i]->Serialize(this, true, info);

	WriteCloseTag("call");
	WriteSeparator();

	return EndSerialization();
	}

bool Serializer::Serialize(const StateAccess& s)
	{
	if ( ! StartSerialization("state access", 's') )
		return false;

	if ( ! s.Serialize(this) )
		{
		Error("failed");
		return false;
		}

	return EndSerialization();
	}

bool Serializer::Serialize(const Timer& t)
	{
	if ( ! StartSerialization("timer", 't') )
		return false;

	if ( ! t.Serialize(this) )
		{
		Error("failed");
		return false;
		}

	return EndSerialization();
	}

bool Serializer::Serialize(const Connection& c)
	{
	if ( ! StartSerialization("connection", 'c') )
		return false;

	// current_cache->Clear();
	if ( ! c.Serialize(this) )
		{
		Error("failed");
		return false;
		}

	return EndSerialization();
	}

int Serializer::Unserialize(bool block)
	{
	assert(current_cache);

	SetErrorDescr("unserializing");

	ChunkedIO::Chunk* chunk = 0;

	while ( ! chunk )
		{
		if ( ! io->Read(&chunk) )
			{
			if ( io->Eof() )
				return 0;
			Error(io->Error());
			return -1;
			}

		if ( ! chunk && ! block )
			return 0;
		}

	format->StartRead(chunk->data, chunk->len);

	char type;
	if ( ! format->Read(&type) )
		return false;

	bool result;
	switch ( type ) {
	case 'i':
		result = UnserializeID();
		break;

	case 'e':
		result = UnserializeCall();
		break;

	case 's':
		result = UnserializeStateAccess();
		break;

	case 'c':
		result = UnserializeConnection();
		break;

	case 't':
		result = UnserializeTimer();
		break;

	default:
		Error(fmt("unknown serialization type %x", (int) type));
		result = false;
	}

	format->EndRead();

	delete [] chunk->data;
	delete chunk;

	return result;
	}

bool Serializer::UnserializeID()
	{
	SetErrorDescr("unserializing ID");

	ID* id = ID::Unserialize(this, false);

	if ( ! id )
		return false;

	if ( verbose )
		{
		ODesc d;
		d.SetQuotes(true);
		id->Describe(&d);
		fprintf(stderr, "Got ID %s\n", d.Description());
		}

	GotID(id, id->ID_Val());

	return true;
	}

bool Serializer::UnserializeCall()
	{
	char* name;
	int len;
	double time;

	if ( ! (Read(&name, 0) && Read(&time) && Read(&len)) )
		return false;

	SetErrorDescr(fmt("unserializing event/function %s", name));

	bool ignore = false;
	FuncType* functype = 0;
	type_list* types = 0;

	ID* id = global_scope()->Lookup(name);

	if ( id )
		{
		if ( id->Type()->Tag() == TYPE_FUNC )
			{
			functype = id->Type()->AsFuncType();
			types = functype->ArgTypes()->Types();
			if ( types->length() != len )
				{
				Error("wrong number of arguments, ignoring");
				ignore = true;
				}
			}
		else
			{
			Error("not a function/event, ignoring");
			ignore = true;
			}
		}
	else
		{
		Error("unknown event/function, ignoring");
		ignore = true;
		}

	ODesc d;
	d.SetQuotes(true);

	val_list* args = new val_list;
	for ( int i = 0; i < len; ++i )
		{
		Val* v = Val::Unserialize(this, 0, TYPE_ANY);

		if ( ! v )
			{
			delete [] name;
			return false;
			}

		if ( v->Type()->Tag() != (*types)[i]->Tag() && ! ignore )
			{
			Error("mismatch in argument types; ignoring");
			ignore = true;
			}

		if ( verbose && types && ! ignore )
			v->Describe(&d);

		args->append(v);
		}

	if ( ! ignore  )
		{
		if ( verbose )
			fprintf(stderr, "%s [%.06f] %s(%s)\n",
				functype->IsEvent() ? "event" : "function call",
				time, name, types ? d.Description() : "<ignored>");

		if ( functype->IsEvent() )
			{
			EventHandler* handler = event_registry->Lookup(name);
			assert(handler);
			GotEvent(name, time, handler, args);
			}
		else
			GotFunctionCall(name, time, id->ID_Val()->AsFunc(), args);
		}
	else
		delete_vals(args);

	delete [] name;

	return true;
	}

bool Serializer::UnserializeStateAccess()
	{
	SetErrorDescr("unserializing state acess");

	StateAccess* s = StateAccess::Unserialize(this);

	if ( ! s )
		return false;

	GotStateAccess(s);

	if ( verbose )
		{
		ODesc d;
		d.SetQuotes(true);
		s->Describe(&d);
		fprintf(stderr, "Got state access: %s\n", d.Description());
		}

	return true;
	}

bool Serializer::UnserializeTimer()
	{
	SetErrorDescr("unserializing timer");

	Timer* t = Timer::Unserialize(this);

	if ( ! t )
		return false;

	GotTimer(t);

	if ( verbose )
		{
		ODesc d;
		d.SetQuotes(true);
		t->Describe(&d);
		fprintf(stderr, "Got timer: %s\n", d.Description());
		}

	return true;
	}

bool Serializer::UnserializeConnection()
	{
	SetErrorDescr("unserializing connection");

	// current_cache->Clear();
	Connection* c = Connection::Unserialize(this);

	if ( ! c )
		return false;

	GotConnection(c);

	if ( verbose )
		{
		ODesc d;
		d.SetQuotes(true);
		c->Describe(&d);
		fprintf(stderr, "Got connection: %s", d.Description());
		}

	return true;
	}

void Serializer::Error(const char* str)
	{
	char buffer[1024];
	snprintf(buffer, sizeof(buffer), "%s%s%s",
		error_descr ? error_descr : "", error_descr ? ": " : "", str);
	ReportError(buffer);
	}

void Serializer::Warning(const char* str)
	{
	char buffer[1024];
	snprintf(buffer, sizeof(buffer), "%s%s%s\n",
		error_descr ? error_descr : "", error_descr ? ": " : "", str);
	fputs(buffer, stderr);
	}

SerializationCache::SerializationCache()
	{
	next_id = 1;
	object_map = new ObjectMap;
	}

SerializationCache::~SerializationCache()
	{
	Clear();
	delete object_map;
	}

SerializationCache::PermanentID SerializationCache::Register(const BroObj* obj, const TransientID& tid, PermanentID pid)
	{
	pid = Register((const SerialObj*) obj, tid, pid);
	id_table[pid].second = true;
	Ref(const_cast<BroObj*>(obj));
	return pid;
	}

SerializationCache::PermanentID SerializationCache::Register(const SerialObj* obj, const TransientID& tid, PermanentID pid)
	{
	if ( pid == NONE )
		pid = next_id++;

	HashKey key(tid.Value());

	delete (PermanentID*) object_map->Remove(&key);

	object_map->Insert(&key, new PermanentID(pid));

	if ( pid >= id_table.size() )
		id_table.resize(pid+1);

	else if ( id_table[pid].second )
		Unref((BroObj*) id_table[pid].first);

	id_table[pid].first = obj;
	id_table[pid].second = false;

	return pid;
	}

void SerializationCache::Clear()
	{
	PermanentID* pid;
	IterCookie* cookie = object_map->InitForIteration();
	while ( (pid = object_map->NextEntry(cookie)) )
		delete pid;

	for ( unsigned int i = 0; i < id_table.size(); i++ )
		if ( id_table[i].second )
			Unref((BroObj*) id_table[i].first);

	next_id = 1;
	id_table.clear();

	delete object_map;
	object_map = new ObjectMap;
	}

FileSerializer::FileSerializer(SerializationFormat* format): Serializer(format)
	{
	file = 0;
	fd = -1;
	io = 0;
	SetCache(&cache);
	}

FileSerializer::~FileSerializer()
	{
	if ( io )
		io->Flush();

	delete [] file;
	delete io;

	if ( fd >= 0 )
		close(fd);
	}

bool FileSerializer::Open(const char* file, bool pure)
	{
	if ( ! OpenFile(file, false) )
		return false;

	if ( pure )
		io->MakePure();

	if ( ! PrepareForWriting() )
		return false;

	return true;
	}

bool FileSerializer::Close()
	{
	CloseFile();
	return true;
	}

bool FileSerializer::OpenFile(const char* arg_file, bool readonly)
	{
	CloseFile();

	cache.Clear();

	file = copy_string(arg_file);
	fd = open(file, readonly ? O_RDONLY : O_WRONLY | O_CREAT | O_TRUNC, 0600);

	if ( fd < 0 )
		{
		if ( readonly && errno == ENOENT )
			{
			// Only an error if we want to print it.
			if ( IsPrintOnly() )
				{
				Error(fmt("%s does not exist", file));
				return false;
				}

			CloseFile();
			return true;
			}

		Error(fmt("can't open file %s for %s: %s",
				file, (readonly ? "reading" : "writing"),
				strerror(errno)));
		return false;
		}

	io = new ChunkedIOFd(fd);

	return io != 0;
	}

void FileSerializer::CloseFile()
	{
	if ( io )
		io->Flush();

	if ( fd >= 0 )
		close(fd);
	fd = -1;

	delete [] file;
	file = 0;

	delete io;
	io = 0;
	}

bool FileSerializer::PrepareForWriting()
	{
	if ( ! io->IsPure() )
		{
		// Write file header.
		uint32 magic = htonl(MAGIC);
		uint16 version = htons(DATA_FORMAT_VERSION);
		uint32 time = htonl((uint32)::time(0));

		if ( write(fd, &magic, sizeof(magic)) != sizeof(magic ) ||
		     write(fd, &version, sizeof(version)) != sizeof(version) ||
		     write(fd, &time, sizeof(time)) != sizeof(time))
			{
			Error(fmt("can't write file header to %s: %s",
					file, strerror(errno)));
			return false;
			}
		}

	return true;
	}

bool FileSerializer::ReadHeader()
	{
	uint32 magic;
	uint16 version;
	uint32 time;

	if ( read(fd, &magic, sizeof(magic)) != sizeof(magic ) ||
	     read(fd, &version, sizeof(version)) != sizeof(version) ||
	     read(fd, &time, sizeof(time)) != sizeof(time) )
		{
		Error(fmt("can't read file header from %s: %s",
				file, strerror(errno)));
		return false;
		}

	version = ntohs(version);
	time = ntohl(time);

	if ( verbose )
		{
		time_t teatime = (time_t) time;
		fprintf(stderr, "Date: %s", ctime(&teatime));
		}

	if ( magic != htonl(MAGIC) )
		{
		Error(fmt("%s is not a bro state file", file));
		CloseFile();
		return false;
		}

	if ( version != DATA_FORMAT_VERSION )
		{
		Error(fmt("wrong data format, expected version %d but got version %d", DATA_FORMAT_VERSION, version));
		CloseFile();
		return false;
		}

	return true;
	}

bool FileSerializer::Read(const char* file)
	{
	if ( ! OpenFile(file, true) )
		return false;

	// fprintf( stderr, "Reading %s\n", file );

	if ( fd < 0 )
		// Not existent, but that's ok.
		return true;

	if ( ! ReadHeader() )
		return false;

	int i;
	while ( (i = Unserialize(true)) > 0 )
		;

	CloseFile();

	return i == 0;
	}

void FileSerializer::ReportError(const char* str)
	{
	run_time(str);
	}

void FileSerializer::GotID(ID* id, Val* val)
	{
	ODesc d;
	d.SetQuotes(true);

	if ( id->AsType() )
		id->AsType()->Describe(&d);
	else if ( val )
		val->Describe(&d);
	else
		d.Add("<nil>");

	fprintf(stdout, "ID %s = %s\n", id->Name(), d.Description());
	}

void FileSerializer::GotStateAccess(StateAccess* s)
	{
	ODesc d;
	d.SetQuotes(true);
	s->Describe(&d);
	fprintf(stdout, "StateAccess %s\n", d.Description());
	}

void FileSerializer::GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args)
	{
	ODesc d;
	d.SetQuotes(true);
	for ( int i = 0; i < args->length(); i++ )
		(*args)[i]->Describe(&d);

	fprintf(stdout, "Event [%.06f] %s(%s)\n", time, name, d.Description());
	}

void FileSerializer::GotFunctionCall(const char* name, double time,
				Func* func, val_list* args)
	{
	ODesc d;
	d.SetQuotes(true);
	for ( int i = 0; i < args->length(); i++ )
		(*args)[i]->Describe(&d);

	fprintf(stdout, "Function call [%.06f] %s(%s)\n", time, name, d.Description());
	}

void FileSerializer::GotTimer(Timer* t)
	{
	ODesc d;
	d.SetQuotes(true);
	t->Describe(&d);
	fprintf(stdout, "Timer %s\n", d.Description());
	}

void FileSerializer::GotConnection(Connection* c)
	{
	ODesc d;
	d.SetQuotes(true);
	c->Describe(&d);
	fprintf(stdout, "Connection %s", d.Description());
	}

ConversionSerializer::ConversionSerializer(SerializationFormat* in,
						SerializationFormat* out)
: FileSerializer(in)
	{
	serout = new FileSerializer(out);
	}

ConversionSerializer::~ConversionSerializer()
	{
	delete serout;
	}

bool ConversionSerializer::Convert(const char* file_in, const char* file_out)
	{
	if ( ! serout->Open(file_out, true) )
		return false;

	if ( ! Read(file_in) )
		return false;

	if ( ! serout->Close() )
		return false;

	return true;
	}

void ConversionSerializer::GotEvent(const char* name, double time,
					EventHandlerPtr event, val_list* args)
	{
	serout->Serialize(name, args);
	}

void ConversionSerializer::GotFunctionCall(const char* name, double time,
					Func* func, val_list* args)
	{
	serout->Serialize(name, args);
	}

void ConversionSerializer::GotID(ID* id, Val* val)
	{
	warn("ConversionSerializer::GotID not implemented");
	}

void ConversionSerializer::GotStateAccess(StateAccess* s)
	{
	warn("ConversionSerializer::GotID not implemented");
	}

class IncrementalWriteTimer : public Timer {
public:
	IncrementalWriteTimer(double t) : Timer(t, TIMER_INCREMENTAL_WRITE)	{}
	virtual void Dispatch(double t, int is_expire)
		{ persistence_serializer->Proceed(!is_expire); }
};

PersistenceSerializer::PersistenceSerializer()
	{
	serial_running = SERIAL_NONE;
	dir = 0;
	current_id = -1;
	clear_list_when_finished = false;
	serial_ids = 0;
	serial_dir = 0;
	current_conn = 0;
	remote_host = 0;
	remote_port = 0;
	}

PersistenceSerializer::~PersistenceSerializer()
	{
	loop_over_list(persistent_ids, i)
		{
		Unref(persistent_ids[i]->id);
		delete [] persistent_ids[i]->file;
		delete persistent_ids[i];
		}

	delete [] dir;
	}

void PersistenceSerializer::Register(ID* id)
	{
	if ( id->Type()->Tag() == TYPE_FUNC )
		{
		Error("can't register functions as persistent ID");
		return;
		}

	loop_over_list(persistent_ids, i)
		if ( persistent_ids[i]->id == id )
			return;

	ID_Info* info = new ID_Info;
	info->id = id;
	Ref(info->id);

	info->file = copy_string(fmt("state.%s", id->Name()));
	info->body = 0;

	persistent_ids.append(info);
	}

void PersistenceSerializer::Register(Connection* conn)
	{
	if ( persistent_conns.Lookup(conn->Key()) )
		return;

	Ref(conn);
	persistent_conns.Insert(conn->Key(), conn);
	}

void PersistenceSerializer::Unregister(Connection* conn)
	{
	Connection* c = (Connection*) persistent_conns.Remove(conn->Key());
	Unref(c);
	}

bool PersistenceSerializer::CheckTimestamp(const char* file)
	{
	struct stat s;
	if ( stat(file, &s) < 0 )
		{
		Error(fmt("can't stat %s: %s", file, strerror(errno)));
		return false;
		}

	if ( ! S_ISREG(s.st_mode) )
		return false;

	bool changed = true;

	HashKey* key = new HashKey(file, strlen(file));
	time_t* t = files.Lookup(key);

	if ( t && *t >= s.st_mtime )
		changed = false;

	if ( ! t )
		files.Insert(key, new time_t(s.st_mtime));
	else
		*t = s.st_mtime;

	delete key;
	return changed;
	}

bool PersistenceSerializer::ReadAll()
	{
	assert(dir);

	struct stat s;
	if ( stat(dir, &s) < 0 )
		{
		if ( errno == ENOENT )
			return true;

		Error(fmt("can't stat %s: %s", dir, strerror(errno)));
		return false;
		}

	DIR* d = opendir(dir);
	if ( ! d )
		{
		Error(fmt("can't open directory %s: %s", dir, strerror(errno)));
		return false;
		}

	dirent* ent;
	while ( (ent = readdir(d)) )
		{
		// Ignore hidden files.
		if ( ent->d_name[0] == '.' )
			continue;

		const char* file = copy_string(fmt("%s/%s", dir, ent->d_name));

		if ( CheckTimestamp(file) )
			{
			Read(file);
			if ( unlink(file) < 0 )
				Error(fmt("can't delete %s: %s",
						file, strerror(errno)));
			}

		delete [] file;
		}

	closedir(d);

	return true;
	}

bool PersistenceSerializer::CheckDir(const char* dir)
	{
	struct stat s;
	if ( stat(dir, &s) < 0 )
		{
		if ( errno != ENOENT )
			{
			Error(fmt("can't stat %s: %s", dir, strerror(errno)));
			return false;
			}

		if ( mkdir(dir, 0700) < 0 )
			{
			Error(fmt("can't create directory %s: %s",
					dir, strerror(errno)));
			return false;
			}
		}

	else if ( ! S_ISDIR(s.st_mode) )
		{
		fprintf( stderr, dir );
		Error(fmt("%s exists but is not a directory", dir));
		return false;
		}

	return true;
	}

bool PersistenceSerializer::CheckSourceFile(const Location& l)
	{
	// Quite a simple heuristic, but should work ...
	// FIXME: There must be a better way than hard-coding this ...
	if ( strstr(l.filename, "/bro.init") ||
	     strstr(l.filename, "/pcap.bro") ||
	     strstr(l.filename, "/const.bif.bro") )
		return false;

	return true;
	}

bool PersistenceSerializer::WriteConfig(const char* dstdir)
	{
	if ( serial_running != SERIAL_NONE )
		return false;

	if ( dstdir )
		bro_logger->Log("Saving config...");

	if ( ! dstdir )
		dstdir = dir;

	if ( ! (CheckDir(dstdir) && CheckDir(fmt("%s/.tmp", dstdir))) )
		return false;

	// Build list of everything we need to store.
	assert(global_ids.length() == 0);
	PDict(ID)* globals = global_scope()->GetIDs();
	IterCookie* c = globals->InitForIteration();

	ID* id;
	while ( (id = globals->NextEntry(c)) )
		{
		if ( id->Type()->Tag() == TYPE_FUNC )
			{
			if ( ! id->HasVal() )
				continue;

			Func* f = id->ID_Val()->AsFunc();

			if ( f->GetKind() == Func::BUILTIN_FUNC )
				continue;	// can't serialize these

			// Split functions with multiple bodies (aka events)
			// into individual entries.
			if ( id->ID_Val()->AsFunc()->IsEvent() )
				{
				const vector<Stmt*>& bodies = f->GetBodies();
				for ( unsigned int j = 0; j < bodies.size(); ++j )
					{
					ID_Info* info = new ID_Info;

					const Location* l =
						bodies[j]->GetLocationInfo();

					if ( ! CheckSourceFile(*l) )
						continue;

					char* tmp = copy_string(l->filename);
					char* base = basename(tmp);
					info->file = copy_string(fmt("event.%s.%s.%d.bst", id->Name(), base, l->first_line));
					delete [] tmp;

					info->body = j;
					info->id = id;
					Ref(info->id);

					global_ids.append(info);
					}
				continue;
				}
			}

		if ( ! CheckSourceFile(*id->GetLocationInfo()) )
			continue;

		ID_Info* info = new ID_Info;
		info->id = id;
		Ref(info->id);

		const char* tag = "id";

		if ( id->AsType() )
			tag = "type";
		else if ( id->Type()->Tag() == TYPE_FUNC )
			tag = "func";

		info->file = copy_string(fmt("%s.%s.bst", tag, id->Name()));
		info->body = 0;	// 0 for functions that aren't events.

		global_ids.append(info);
		}

	no_content = dstdir == dir;
	serial_dir = dstdir;
	serial_ids = &global_ids;
	current_id = 0;
	id_suspended = false;
	clear_list_when_finished = true;

	return Proceed(false);
	}

// FIXME: The following methods are getting more and more ugly. I should
// really clean them up eventually...

bool PersistenceSerializer::SendState(addr_type host, uint16 port)
	{
	if ( serial_running != SERIAL_NONE )
		return false; // still a serialization running.

	remote_host = host;
	remote_port = port;
	return SerializeState(true, "Sending state...");
	}

bool PersistenceSerializer::WriteState(bool incremental)
	{
	if ( serial_running != SERIAL_NONE )
		return false; // still a serialization running.

	remote_host = 0;
	remote_port = 0;
	return SerializeState(incremental, "Checkpointing state...");
	}

bool PersistenceSerializer::SerializeState(bool incremental, const char* log)
	{
	assert(dir);

	if ( ! persistent_ids.length() && ! persistent_conns.Length() )
		{
		if ( remote_host )
			RaiseFinishedSendState();
		return true;
		}

	bro_logger->Log(log);

	if ( ! (CheckDir(dir) && CheckDir(fmt("%s/.tmp", dir))) )
		return false;

	no_content = false;
	serial_dir = dir;
	current_id = 0;
	serial_ids = &persistent_ids;
	id_suspended = false;
	clear_list_when_finished = false;

	current_conn = 0;

	serial_running = persistent_conns.Length() ? SERIAL_CONNS : SERIAL_IDS;

	return Proceed(incremental);
	}

bool PersistenceSerializer::Proceed(bool incremental)
	{
	// First, we serialize all data into the directory
	// <serial-dir>/.tmp/. Then we move the files one level
	// up. This makes sure that the old state keeps consistent
	// if something goes wrong.
	bool success = false;

	// Serialize connections.
	if ( serial_running == SERIAL_CONNS )
		{
		if ( ! current_conn )
			{
			if ( ! remote_host )
				{
				if ( ! OpenFile(fmt("%s/.tmp/connections.bst", serial_dir), false) )
					goto cleanup;

				if ( ! PrepareForWriting() )
					goto cleanup;
				}

			current_conn = persistent_conns.InitForIteration();
			persistent_conns.MakeRobustCookie(current_conn);
			}

		Connection* c;
		while ( (c = persistent_conns.NextEntry(current_conn)) )
			{
			if ( ! Save(*c) )
				goto cleanup;

			if ( incremental )
				{
				timer_mgr->Add(new IncrementalWriteTimer(network_time + state_write_delay));
				return true;
				}
			}

		serial_running = SERIAL_IDS;
		}

	// Serialize IDs.
	while ( current_id < serial_ids->length() )
		{
		ID_Info* id = (*serial_ids)[current_id];

		if ( ! id_suspended )
			{
			if ( ! remote_host )
				{
				if ( ! OpenFile(fmt("%s/.tmp/%s",
						serial_dir, id->file), false) )
					goto cleanup;

				if ( ! PrepareForWriting() )
					goto cleanup;
				}

			++current_id;
			}

		if ( ! no_content )
			{
			SerialInfo info(incremental ?  &id_suspended : 0,
					id_suspended, id->body);

			if ( ! Save(*id->id, &info) )
				goto cleanup;

			if ( info.incr &&
			     (current_id < serial_ids->length() ||
			      id_suspended) )
				{
				timer_mgr->Add(new IncrementalWriteTimer(network_time + state_write_delay));
				return true;
				}
			}
		}

	if ( remote_host )
		{
		RaiseFinishedSendState();
		success = true;
		goto cleanup;
		}

	CloseFile();

	// Before we move the files to their new place, we first check if their
	// is any new data. If so, we read it in before overwriting it. When we do
	// the next checkpointing, the new data will be written.
	//
	// Yes, there's a race condition here.
	ReadAll();

	char oldname[PATH_MAX];
	char newname[PATH_MAX];

	char* oldp;
	char* newp;

	oldp = oldname + snprintf(oldname, PATH_MAX, "%s/.tmp/", serial_dir );
	newp = newname + snprintf(newname, PATH_MAX, "%s/", serial_dir );

	// If a tmp/connections.bst exists, move it up.
	strncpy(oldp, "connections.bst", PATH_MAX - (oldp - oldname));
	oldname[PATH_MAX-1] = '\0';

	struct stat dummy;
	if ( stat(oldname, &dummy) >= 0 )
		{
		strncpy(newp, "connections.bst", PATH_MAX - (newp - newname));
		newname[PATH_MAX-1] = '\0';
		if ( rename(oldname, newname) < 0 )
			{
			Error(fmt("can't move %s to %s: %s", oldname, newname,
					  strerror(errno)));
			goto cleanup;
			}

		// Update the saved time-stamp.
		CheckTimestamp(newname);
		}

	success = true;

	loop_over_list(*serial_ids, i)
		{
		strncpy(oldp, (*serial_ids)[i]->file,
				PATH_MAX - (oldp - oldname) );
		strncpy(newp, (*serial_ids)[i]->file,
				PATH_MAX - (newp - newname) );

		oldname[PATH_MAX-1] = newname[PATH_MAX-1] = '\0';

		if ( rename(oldname, newname) < 0 )
			{
			Error(fmt("can't move %s to %s: %s",
					oldname, newname, strerror(errno)));
			success = false;
			}

		// Update the saved time-stamp.
		CheckTimestamp(newname);
		}

	if ( ! no_content )
		bro_logger->Log("Serialization done.");

cleanup:
	serial_running = SERIAL_NONE;

	if ( clear_list_when_finished )
		{
		loop_over_list(*serial_ids, i)
			{
			Unref((*serial_ids)[i]->id);
			delete [] (*serial_ids)[i]->file;
			delete (*serial_ids)[i];
			}

		serial_ids->clear();
		serial_ids = 0;
		}

	if ( ! success )
		bro_logger->Log("Serialization error.");

	return success;
	}


void PersistenceSerializer::RaiseFinishedSendState()
	{
	val_list* vl = new val_list;
	vl->append(new AddrVal(htonl(remote_host)));
	vl->append(new PortVal(remote_port));

	mgr.QueueEvent(finished_send_state, vl);
	bro_logger->Log("Serialization done.");
	}

bool PersistenceSerializer::Save(const Connection& c)
	{
	if ( ! c.Key() )
		// This connection has been removed in the meantime.
		return true;

	if ( remote_host )
		return remote_serializer->SendConnection(remote_host, remote_port, c);

	return Serialize(c);
	}

bool PersistenceSerializer::Save(const ID& id, SerialInfo* info)
	{
	if ( remote_host )
		{
		// No incremental serialization here yet. It's not clear
		// if we need it anyway.
		return remote_serializer->SendID(remote_host, remote_port, id);
		}

	return Serialize(id, *info);
	}

void PersistenceSerializer::GotEvent(const char* name, double time,
					EventHandlerPtr event, val_list* args)
	{
	mgr.QueueEvent(event, args);
	}

void PersistenceSerializer::GotFunctionCall(const char* name, double time,
					Func* func, val_list* args)
	{
	func->Call(args);
	}

void PersistenceSerializer::GotStateAccess(StateAccess* s)
	{
	run_time("PersistenceSerializer::GotStateAccess not implemented");
	}

void PersistenceSerializer::GotTimer(Timer* s)
	{
	run_time("PersistenceSerializer::GotTimer not implemented");
	}

void PersistenceSerializer::GotConnection(Connection* c)
	{
	// Install connection.
	if ( c->IsPersistent() && c->Key() )
		persistence_serializer->Register(c);

	sessions->Insert(c);
	}

void PersistenceSerializer::GotID(ID* id, Val* val)
	{
	// Install the ID.
	if ( ! id->IsGlobal() )
		{
		Error("persistent non-global id");
		delete id;
		return;
		}

	ID* current = global_scope()->Lookup(id->Name());

	if ( ! current )
		{
		global_scope()->Insert(id->Name(), id);
		if ( id->FindAttr(ATTR_PERSISTENT) )
			Register(id);
		return;
		}

	if ( ! val )
		{
		Error("type already exists; ignored.");
		return;
		}

	// If it's an event handler for an already existing event, add
	// this to the current list of bodies.
	if ( val->Type()->Tag() == TYPE_FUNC &&
	     val->Type()->AsFuncType()->IsEvent() && current->HasVal() )
		{
		assert(val->AsFunc()->GetKind() == Func::BRO_FUNC);
		BroFunc* f = (BroFunc*) val->AsFunc();
		Func* cur = current->ID_Val()->AsFunc();
		const vector<Stmt*>& bodies = f->GetBodies();

		for( unsigned int i = 0; i < bodies.size(); i++ )
			cur->AddBody(bodies[i], 0, f->FrameSize());

		Unref(id);
		return;
		}

	// Change the value.
	// Type safety with respect to an existing global ID should
	// have already been assured.
	// FIXME: What about attributes? What about the other members?
	Ref(val);
	current->SetVal(val);

	if ( val && val->Type()->Tag() == TYPE_FUNC && val->AsFunc() )
		// We know that all our functions are in fact debug functions...
		((DebugFunc*) val->AsFunc())->SetID(current);

	Unref(id);
	}

bool PersistenceSerializer::LogAccess(const StateAccess& s)
	{
	internal_error("PersistenceSerializer::LogAccess not implemented");
	}

EventPlayer::EventPlayer()
	{
	stream_time = 0;
	replay_time = 0;
	ne_time = 0;
	}

bool EventPlayer::Replay(const char* file)
	{
	if ( ! OpenFile(file, true) )
		return false;

	if ( fd < 0 )
		{
		Error(fmt("%s does not exist", file));
		return false;
		}

	return ReadHeader();
	}

EventPlayer::~EventPlayer()
	{
	CloseFile();
	}

void EventPlayer::GotEvent(const char* name, double time,
		EventHandlerPtr event, val_list* args)
	{
	ne_time = time;
	ne_handler = event;
	ne_args = args;
	}

void EventPlayer::GotFunctionCall(const char* name, double time,
		Func* func, val_list* args)
	{
	// We don't replay function calls.
	}

double EventPlayer::NextTimestamp()
	{
	if ( ne_time )
		return ne_time;

	if ( ! io )
		return 0;

	// Read next event if we don't have one waiting.
	if ( ! ne_time )
		Unserialize();

	if ( ! ne_time )
		return 0;

	if ( ! network_time )
		{
		// Network time not initialized yet.
		stream_time = replay_time = ne_time;
		return ne_time;
		}

	if ( ! stream_time )
		{
		// Init base times.
		stream_time = ne_time;
		replay_time = network_time;
		}

	// Scale time.
	ne_time = ne_time - stream_time + network_time;
	return ne_time;
	}

void EventPlayer::DispatchNextEvent()
	{
	if ( ! ( io && ne_time ) )
		return;

	Event* event = new Event(ne_handler, ne_args);
	mgr.Dispatch(event);

	ne_time = 0;
	}

bool EventPlayer::IsActive()
	{
	if ( ! io )
		return false;

	return ! io->Eof();
	}
