// $Id: Serializer.h,v 1.4 2004/03/01 12:23:25 vern Exp $

#ifndef SERIALIZER_H
#define SERIALIZER_H

#include "ID.h"
#include "List.h"
#include "Expr.h"
#include "ChunkedIO.h"
#include "SerializationFormat.h"
#include "StateAccess.h"

class SerializationCache;
class SerialInfo;

class Connection;
class Timer;

class Serializer {
public:
	bool Serialize(const ID& id, const SerialInfo& info);
	bool Serialize(const char* func, val_list* args);
	bool Serialize(const StateAccess& s);
	bool Serialize(const Connection& c);
	bool Serialize(const Timer& t);

	// Turns on output during unserialization.
	void Verbose()	{ verbose = 1; }

	// Tells the unserializers that all we want to do is printing the
	// state.  This may make some errors non-fatal.
	//
	// FIXME: The better approach would be to generally pass some
	// UnserialInfo to all Unserialize() functions (and, BTW, we
	// should make the use of SerialInfo more consistent, too.)

	void SetPrintOnly(bool arg_print_only)	{ print_only = arg_print_only; }
	bool IsPrintOnly() const	{ return print_only; }

	// Access to the current cache.
	SerializationCache* Cache()	{ return current_cache; }
	void SetCache(SerializationCache* cache)
		{ current_cache = cache; }

	// Input/output methods.

#define DECLARE_READ(type) \
	bool Read(type* v)	{ return format->Read(v); }

#define DECLARE_WRITE(type) \
	bool Write(type v, const char* tag = 0)	\
		{ return format->Write(v, tag); }

#define DECLARE_IO(type)	\
	DECLARE_READ(type)	\
	DECLARE_WRITE(type)

	DECLARE_IO(int)
	DECLARE_IO(uint16)
	DECLARE_IO(uint32)
	DECLARE_IO(char)
	DECLARE_IO(bool)
	DECLARE_IO(double)

	bool Read(char** str, int* len)
		{ return format->Read(str, len); }
	bool Read(const char** str, int* len)
		// This cast is ok.
		{ return format->Read(const_cast<char**>(str), len); }

	bool Read(string* s);

	bool Write(const char* s, const char* tag = 0)
		{ return format->Write(s, tag); }
	bool Write(const char* buf, int len, const char* tag = 0)
		{ return format->Write(buf, len, tag); }
	bool Write(const string& s, const char* tag = 0)
		{ return format->Write(s.data(), s.size(), tag); }

	bool WriteOpenTag(const char* tag)
		{ return format->WriteOpenTag(tag); }
	bool WriteCloseTag(const char* tag)
		{ return format->WriteCloseTag(tag); }

	bool WriteSeparator()	{ return format->WriteSeparator(); }

	void Error(const char* msg);
	void Warning(const char* msg);

	void SetErrorDescr(const char* descr)
		{ delete [] error_descr; error_descr = copy_string(descr); }

protected:
	// Format defaults to binary serialization.
	Serializer(SerializationFormat* format = 0);
	virtual ~Serializer();

	// Reads next object.
	// If 'block' is true, wait until an object can be read.
	// Returns 0 if no more object available, -1 on error.
	int Unserialize(bool block = false);

	// Callback for error messages.
	virtual void ReportError(const char* msg) = 0;

	// Callbacks for unserialized objects.

	// id points to ID in global scope, val is unserialized value.
	virtual void GotID(ID* id, Val* val) = 0;
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args) = 0;
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args) = 0;
	virtual void GotStateAccess(StateAccess* s) = 0;
	virtual void GotTimer(Timer* t) = 0;
	virtual void GotConnection(Connection* c) = 0;

	// Magic to recognize state files.
	static const uint32 MAGIC = 0x42525354;

	// This will be increased whenever there is an incompatible change
	// in the data format.
	static const int DATA_FORMAT_VERSION = 7;

	bool verbose;
	ChunkedIO* io;

private:
	bool StartSerialization(const char* descr, char tag);
	bool EndSerialization();

	bool UnserializeID();
	bool UnserializeCall();
	bool UnserializeStateAccess();
	bool UnserializeTimer();
	bool UnserializeConnection();

	SerializationFormat* format;
	SerializationCache* current_cache;
	const char* error_descr;	// used in error messages

	bool print_only;
};



// We maintain a cache for some of the objects which have already been
// serialized. For the cache, we need two types of IDs: TransientIDs (defined
// in SerialObj.cc) uniquely reference an object during the lifetime of a
// process.  PermanentIDs uniquely reference an object within a
// serialization.

class SerializationCache {
public:
	typedef uint16 PermanentID;
	static const PermanentID NONE = 0;

	SerializationCache();
	~SerializationCache();

	PermanentID Register(const BroObj* obj, const TransientID& tid,
				PermanentID pid);
	PermanentID Register(const SerialObj* obj, const TransientID& tid,
				PermanentID pid = NONE);

	const SerialObj* Lookup(PermanentID pid) const
		{ return pid < id_table.size() ? id_table[pid].first : 0; }

	PermanentID Lookup(const TransientID& tid) const
		{
		HashKey key(tid.Value());
		PermanentID* id = object_map->Lookup(&key);
		return id ? *id : 0;
		}

	void Clear();

private:
	// Object cache.
	declare(PDict, PermanentID);
	typedef PDict(PermanentID) ObjectMap;
	// bool is true if it's an BroObj.
	typedef pair<const SerialObj*, bool> id_entry;
	typedef vector<id_entry> IDTable;

	IDTable id_table;
	ObjectMap* object_map;
	PermanentID next_id;
};

// Data to pass on to the serialization functions.
// (we have to declare it outside Serializer to make it
// forward-declarable... :-( ).
struct SerialInfo {
	SerialInfo(bool* arg_incr = 0, bool arg_cont = false, int arg_body = 0)
		{ incr = arg_incr; cont = arg_cont; body = arg_body; }

	// If non-nil, suspending the serialization is ok. Then
	// *incr has to be set to true to signal that the serialization
	// has not been finished yet.
	bool* incr;

	// If true, we're continuing a previously suspended serialization.
	bool cont;

	// If we serialize a function with multiple bodies (aka event handler),
	// this gives the number of the body we're refering to.
	int body;
};

// Write values/events to file or fd.
class FileSerializer : public Serializer {
public:
	FileSerializer(SerializationFormat* format = 0);
	virtual ~FileSerializer();

	// Opens the file for serialization.
	bool Open(const char* file, bool pure = false);
	bool Close();

	// Reads the file and prints content to stderr.
	bool Read(const char* file);

protected:
	virtual void ReportError(const char* msg);
	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args);
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args);
	virtual void GotStateAccess(StateAccess* s);
	void GotTimer(Timer* t);
	void GotConnection(Connection* c);

	bool OpenFile(const char* file, bool readonly);
	void CloseFile();
	bool ReadFile(const char* file);
	bool PrepareForWriting();
	bool ReadHeader();

	SerializationCache cache;
	const char* file;
	int fd;
};

// Converts from one serialization format into another.
class ConversionSerializer:public FileSerializer {
public:
	ConversionSerializer(SerializationFormat* in, SerializationFormat* out);
	virtual ~ConversionSerializer();

	bool Convert(const char* file_in, const char* file_out);

protected:
	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args);
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args);
	virtual void GotStateAccess(StateAccess* s);

	FileSerializer* serout;
};


// Implements persistance for values.
class PersistenceSerializer : public FileSerializer {
public:
	PersistenceSerializer();
	virtual ~PersistenceSerializer();

	// Define the directory where the data is to be stored.
	void SetDir(const char* arg_dir)	{ dir = copy_string(arg_dir); }

	// Registers the ID/connection to be saved by WriteAll().
	void Register(ID* id);
	void Register(Connection* conn);
	void Unregister(Connection* conn);

	// Tells the serializer that this access was performed. If a
	// serialization is going on, it may store it.
	bool LogAccess(const StateAccess& s);

	// Reads all data that has been changed since last scan.
	bool ReadAll();

	// Writes all registered data. Existing data in
	// dir is overwritten. If incremental is true, may suspend
	// serialization after some amount of data is written.
	bool WriteState(bool incremental);

	// Sends all registered state to remote host
	// (by leveraging the remote_serializer).
	bool SendState(addr_type host, uint16 port);

	// Writes Bro's configuration (w/o dynamic state).
	bool WriteConfig(const char* dstdir);

	// Returns true if last (incremental) WriteAll() hasn't finished yet.
	bool InProgress()	{ return serial_running != SERIAL_NONE; }

	// Performs (one step of incremental) serialization.
	bool Proceed(bool incremental);

protected:
	friend class RemoteSerializer;

	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args);
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args) ;
	virtual void GotStateAccess(StateAccess* s);
	virtual void GotTimer(Timer* t);
	virtual void GotConnection(Connection* c);

	// Make sure directory exists (creating it if not).
	bool CheckDir(const char* dir);

	// Checks if we want to store things from this location
	// (they may just clutter up our output).
	bool CheckSourceFile(const Location& l);

	// Returns true if it's a regular file and has a more recent timestamp
	// than last time we checked it.
	bool CheckTimestamp(const char* file);

	// Generates an error message, terminates current serialization,
	// and returns false.
	bool SerialError(const char* msg);

	// Main initialization method for serialization.
	bool SerializeState(bool incremental, const char* log);

	// Wrappers which either serializes to disk or sends to remote host.
	bool Save(const Connection& c);
	bool Save(const ID& i, SerialInfo* info);

	// Sends event when state has been sent completely.
	void RaiseFinishedSendState();

private:
	// Main state directory.
	const char* dir;

	struct ID_Info {
		ID* id;
		const char* file;	// File name w/o path.

		// For events: the body we want to reference.
		unsigned int body;
	};

	declare(PList, ID_Info);
	typedef PList(ID_Info) id_info_list;

	declare(PDict, Connection);
	typedef PDict(Connection) conn_map;

	// Registered state IDs to store.
	id_info_list persistent_ids;

	// Registered connections to store.
	conn_map persistent_conns;

	// Temporary list of global IDs to store.
	id_info_list global_ids;

	// To keep track of files' modification times.
	declare(PDict, time_t);
	typedef PDict(time_t) file_map;
	file_map files;

	// State for incremental serialization.
	enum { SERIAL_NONE, SERIAL_IDS, SERIAL_CONNS } serial_running;
	bool no_content;
	bool id_suspended;
	bool clear_list_when_finished;
	int current_id;
	id_info_list* serial_ids;
	IterCookie* current_conn;
	addr_type remote_host;
	uint16 remote_port;

	const char* serial_dir;
};


// Abstract interface class for external sources providing a stream of events.
class EventSource {
public:
	// Returns time of the oldest event (0 if none available).
	virtual double NextTimestamp() = 0;

	// Dispatches the oldest event and removes it.
	virtual void DispatchNextEvent() = 0;

	// Returns true if there are more events to expect from this source.
	virtual bool IsActive() = 0;
};

// Plays a file of events back (not implemeted yet).
class EventPlayer : public FileSerializer, public EventSource {
public:
	EventPlayer();
	virtual ~EventPlayer();

	bool Replay(const char* file);

	virtual double NextTimestamp();
	virtual void DispatchNextEvent();
	virtual bool IsActive();

protected:
	virtual void GotID(ID* id, Val* val)	{}
	virtual void GotEvent(const char* name, double time,
				EventHandlerPtr event, val_list* args);
	virtual void GotFunctionCall(const char* name, double time,
				Func* func, val_list* args);

	double stream_time;	// time of first captured event
	double replay_time;	// network time of replay start

	// Next event waiting to be dispatched.
	double ne_time;
	EventHandlerPtr ne_handler;
	val_list* ne_args;

};

extern EventPlayer* event_player;
extern FileSerializer* event_serializer;
extern PersistenceSerializer* persistence_serializer;

declare(PList, EventSource);
typedef PList(EventSource) event_source_list;
extern event_source_list external_event_sources;


#endif
