// $Id: Stats.cc,v 1.5 2004/03/01 12:23:25 vern Exp $

#include "Conn.h"
#include "File.h"
#include "Event.h"
#include "NetVar.h"
#include "Sessions.h"
#include "Stats.h"
#include "Scope.h"
#include "cq.h"

#ifdef HAVE_MALLINFO
# include <malloc.h>
#endif

class StatsTimer : public Timer {
public:
	StatsTimer(double t, StatsLogger* l, double i) : Timer(t, TIMER_STATS)
		{
		logger = l;
		interval = i;
		}

	virtual void Dispatch(double t, int is_expire)
		{
		if ( active )
			logger->Log();

		// Reinstall timer.
		if ( ! is_expire )
			timer_mgr->Add(new StatsTimer(network_time + interval,
							logger, interval));
		}

protected:
	double interval;
	StatsLogger* logger;
};

StatsLogger::StatsLogger(BroFile* arg_file, double interval)
	{
	file = arg_file;
	timer_mgr->Add(new StatsTimer(1, this, interval));
	}

StatsLogger::~StatsLogger()
	{
	file->Close();
	}

void StatsLogger::Log()
	{
	file->Write(fmt("%.06f ------------------------\n", network_time));

	// Memory information.
#ifdef HAVE_MALLINFO
	struct rusage r;
	getrusage(RUSAGE_SELF, &r);
	struct timeval tv_utime = r.ru_utime;
	struct timeval tv_stime = r.ru_stime;
	
	// For memory, getrusage() gives bogus results on Linux. Grmpf.
	const struct mallinfo& mi = mallinfo();
	unsigned int malloced = mi.uordblks;
	unsigned int total = mi.arena;
#else
	struct rusage r;
	getrusage(RUSAGE_SELF, &r);
	unsigned int malloced = 0;
	unsigned int total = r.ru_maxrss * 1024; // at least on FreeBSD it's in KB
	struct timeval tv_utime = r.ru_utime;
	struct timeval tv_stime = r.ru_stime;
#endif

	static unsigned int first_total = 0;
	static double first_rtime = 0;
	static double first_utime = 0;
	static double first_stime = 0;

	double rtime = current_time();
	double utime = double(tv_utime.tv_sec) + double(tv_utime.tv_usec) / 1e6;
	double stime = double(tv_stime.tv_sec) + double(tv_stime.tv_usec) / 1e6;

	if ( first_total == 0 )
		{
		first_total = total;
		first_rtime = rtime;
		first_utime = utime;
		first_stime = stime;
		
		file->Write(fmt("%.06f Command line: ", network_time ));
		for ( int i = 0; i < bro_argc; i++ )
			{
			file->Write(bro_argv[i]);
			file->Write(" ");
			}
		file->Write(fmt("\n%.06f ------------------------\n", network_time));
		}
	
	file->Write(fmt("%.06f Mem: total=%dK total_adj=%dK malloced: %dK\n",
		network_time, total / 1024, (total - first_total) / 1024,
		malloced / 1024));
		
	// Run-time
	file->Write(fmt("%.06f Run-time: user+sys=%.1f user=%.1f sys=%.1f real=%.1f\n",
		network_time, (utime + stime) - (first_utime + first_stime),
		utime - first_utime, stime - first_stime, rtime - first_rtime));
	
	// Connection information.
	file->Write(fmt("%.06f Conns: total=%d current=%d/%d mem=%dK avg=%.1f table=%dK connvals=%dK\n",
		network_time,
		Connection::TotalConnections(),
		Connection::CurrentConnections(),
		sessions->CurrentConnections(),
		sessions->ConnectionMemoryUsage() / 1024,
		sessions->ConnectionMemoryUsage() /
			double(sessions->CurrentConnections()),
		sessions->MemoryAllocation() / 1024,
		sessions->ConnectionMemoryUsageConnVals() / 1024
		));

	SessionStats s;
	sessions->GetStats(s);
	
	file->Write(fmt("%.06f Conns: tcp=%d/%d udp=%d/%d icmp=%d/%d\n",
		network_time,
		s.num_TCP_conns, s.max_TCP_conns,
		s.num_UDP_conns, s.max_UDP_conns,
		s.num_ICMP_conns, s.max_ICMP_conns
		));

	// Reassembler.
	file->Write(fmt("%.06f Total reassembler data: %dK\n", network_time,
		Reassembler::TotalMemoryAllocation() / 1024));

	// Signature engine.
	if ( rule_matcher )
		{
		RuleMatcher::Stats stats;
		rule_matcher->GetStats(&stats);
		
		file->Write(fmt("%06f RuleMatcher: matchers=%d states=%d computed=%d "
			"mem=%dK\n", network_time, stats.matchers,
			stats.states, stats.computed, stats.mem / 1024));
		}
	
	// Timer information.
	file->Write(fmt("%.06f Timers: current=%d max=%d mem=%dK\n",
		network_time,
		timer_mgr->Size(), timer_mgr->PeakSize(),
		(cq_memory_allocation() +
		 (timer_mgr->Size() * padded_sizeof(ConnectionTimer))) / 1024));
	
	unsigned int* current_timers = TimerMgr::CurrentTimers();
	for( int i = 0; i < NUM_TIMER_TYPES; i++ )
		{
		if ( current_timers[i] )
			file->Write(fmt("%.06f         %s = %d\n", network_time,
					timer_type_to_string((TimerType) i),
					current_timers[i]));
		}

	// Script level state.
	unsigned int size, mem = 0;
	PDict(ID)* globals = global_scope()->Vars();
	IterCookie* c = globals->InitForIteration();

	file->Write(fmt("%.06f Global_sizes > 100k: %dK\n",
			network_time, mem / 1024));

	ID* id;
	while ( (id = globals->NextEntry(c)) )
		if ( id->HasVal() )
			{
			Val* v = id->ID_Val();

			size = id->ID_Val()->MemoryAllocation();
			mem += size;

			if ( size > 100 * 1024 )
				{
				file->Write(fmt("%.06f                %s = %dK",
					network_time, id->Name(), size / 1024));

				if ( v->Type()->Tag() == TYPE_TABLE )
					file->Write(fmt(" (%d entries)\n",
						v->AsTable()->Length()));
				else
					file->Write("\n");
				}
			}

	file->Write(fmt("%.06f Global_sizes total: %dK\n",
			network_time, mem / 1024));

	// Create an event so that scripts can log their information, too.
	// (and for consistency we dispatch it *now*)
	if ( statistics_update )
		{
		val_list* vl = new val_list;
		vl->append(new Val(file));
		mgr.Dispatch(new Event(statistics_update, vl));
		}
	}

void StatsLogger::SegmentProfile(const char* name, const Location* loc,
					double dtime, int dmem)
	{
	if ( name )
		file->Write(fmt("%.06f segment-%s dt=%.06f dmem=%d\n",
				network_time, name, dtime, dmem));
	else if ( loc )
		file->Write(fmt("%.06f segment-%s:%d dt=%.06f dmem=%d\n",
				network_time,
				loc->filename ? loc->filename : "nofile",
				loc->first_line,
				dtime, dmem));
	else
		file->Write(fmt("%.06f segment-XXX dt=%.06f dmem=%d\n",
				network_time, dtime, dmem));
	}


void SegmentProfiler::Init()
	{
	getrusage(RUSAGE_SELF, &initial_rusage);
	}

void SegmentProfiler::Report()
	{
	struct rusage final_rusage;
	getrusage(RUSAGE_SELF, &final_rusage);

	double start_time =
		double(initial_rusage.ru_utime.tv_sec) +
		double(initial_rusage.ru_utime.tv_usec) / 1e6 +
		double(initial_rusage.ru_stime.tv_sec) +
		double(initial_rusage.ru_stime.tv_usec) / 1e6;

	double stop_time =
		double(final_rusage.ru_utime.tv_sec) +
		double(final_rusage.ru_utime.tv_usec) / 1e6 +
		double(final_rusage.ru_stime.tv_sec) +
		double(final_rusage.ru_stime.tv_usec) / 1e6;

	int start_mem = initial_rusage.ru_maxrss * 1024;
	int stop_mem = initial_rusage.ru_maxrss * 1024;

	double dtime = stop_time - start_time;
	int dmem = stop_mem - start_mem;

	logger->SegmentProfile(name, loc, dtime, dmem);
	}
