# $Id: alert.bro,v 1.16 2004/03/15 20:18:35 vern Exp $

type Alert: enum {
	Unspecified,	# place-holder - shouldn't be used
};

type alert_info: record {
	alert: Alert;
	msg: string;
	sub: string &optional;	# sub-message

	conn: connection &optional;	# connection associated with alert
	id: conn_id &optional;	# connection-ID, if we don't have a connection handy
	src: addr &optional;	# source address, if we don't have a connection
	source_is_responder: bool &optional;	# if missing or F, alert source is orig.

	# The following are detailed attributes that are associated with some
	# alerts, but not all.

	user: string &optional;

	filename: string &optional;

	method: string &optional;
	URL: string &optional;

	n: count &optional;	# associated count, or perhaps status code

	# Source that raised the last event before the alert was generated.
	# Automatically set by ALERT().
	event_src: event_source &optional;
};

type AlertAction: enum {
	# Similar to WeirdAction in weird.bro.
	ALERT_IGNORE, ALERT_FILE,
	ALERT_LOG_ALWAYS, ALERT_LOG_PER_CONN,
	ALERT_LOG_PER_ORIG, ALERT_LOG_ONCE,
};

type alert_policy_item: record {
	result: AlertAction;
	pred: function(a: alert_info): bool;
	priority: count;
};

global alert_policy: set[alert_policy_item] = {
	[$pred(a: alert_info) = { return T; },
	 $result = ALERT_LOG_ALWAYS,
	 $priority = 0],
} &redef;

# table that maps alerts into a function that should be called
# to determine the action.
global alert_action_filters:
	table[Alert] of function(a: alert_info): AlertAction &redef;

function ignore_alert(a: alert_info): AlertAction
	{
	return ALERT_IGNORE;
	}

function file_alert(a: alert_info): AlertAction
	{
	return ALERT_FILE;
	}

global alert_tallies: table[string] of count &default = 0;

function tally_alert(s: string)
	{
	++alert_tallies[s];
	}

function tally_alert_type(a: alert_info): AlertAction
	{
	tally_alert(fmt("%s", a$alert));
	return ALERT_FILE;
	}

event bro_done()
	{
	for ( s in alert_tallies )
		{
		local n = alert_tallies[s];
		log fmt("%s (%d time%s)", s, n, n > 1 ? "s" : "");
		}
	}

global alert_file = open_log_file("alert") &redef;

# This handler is useful for processing alerts after the alert filters
# have been applied and yielded an AlertAction.
#
# It's tempting to make the default handler do the logging and
# printing to alert_file, rather than ALERT.  I hesitate to do that,
# though, because it perhaps could slow down notification, because
# in the absence of event priorities, the event would have to wait
# behind any other already-queued events.
event alert_action(a: alert_info, action: AlertAction)
	{
	}

function ALERT(a: alert_info)
	{
	# Fill in some defaults.
	if ( ! a?$id && a?$conn )
		a$id = a$conn$id;

	if ( ! a?$src && a?$id )
		a$src = a$id$orig_h;

	# Event source may already be set if we got this
	# alert from a remote peer.
	if ( ! a?$event_src )
		a$event_src = get_event_source();

	local action = match a using alert_policy;

	if ( a$alert in alert_action_filters )
		action = alert_action_filters[a$alert](a);

	local descr = "";
	if ( action != ALERT_IGNORE && action != ALERT_FILE )
		{
		if ( is_remote_event() )
			{
			if ( a$event_src$descr != "" )
				descr = fmt("<%s> ", a$event_src$descr );
			else
				descr = fmt("<%s/%s> ", a$event_src$host,
							a$event_src$p);
			}

		log fmt("%s %s%s", a$alert, descr, a$msg);
		}

	if ( action != ALERT_IGNORE )
		print alert_file, fmt("%.6f %s %s%s",
					network_time(), a$alert, descr, a$msg);

@ifdef ( IDMEF_support )
	if ( a?$id )
		generate_idmef(a$id$orig_h, a$id$orig_p,
			       a$id$resp_h, a$id$resp_p);
@endif

	event alert_action(a, action);
	}

# function alert(a: alert_info)
# 	{
# 	local id = a?$id ? a$id : a$conn$id;
# 	local addl = a?$conn ? fmt(" (%s)", a$conn$addl) : "";
# 	local source = (a?$source_is_responder && a$source_is_responder) ?
# 			id$resp_h : id$orig_h;
# 	local msg = a?$sub ? fmt("%s %s", a$message, a$sub) : a$message;
#
# 	local log_msg = fmt("%s alert: %s (%s)%s", a$analyzer, msg, id_string(id), addl);
#
# 	log log_msg;
# 	}
