# $Id: hand-over.bro,v 1.1 2004/03/01 12:12:20 vern Exp $
#
# Hand-over between two instances of Bro.

@load remote

# The host from which we want to take over the state has to be
# added to remote_peers_{clear,ssl}, setting hand_over to T.
#
# The host which we want to allow to perform a hand-over with us
# has to be added to remote_peers with a port of 0/tcp and
# hand_over = T.

function is_it_us(host: addr, p: port): bool
	{
@ifdef ( listen_if_clear )
	if ( is_local_interface(host) && p == listen_port_clear )
		return T;
@endif

@ifdef ( listen_if_ssl )
	if ( is_local_interface(host) && p == listen_port_ssl )
	    return T;
@endif
	return F;
	}

function is_handover_peer(ip: addr, p: port): bool
	{
	local peer: Peer;

	if ( [ip, p] in remote_peers_clear )
		peer = remote_peers_clear[ip, p];
	else if ( [ip, p] in remote_peers_ssl )
		peer = remote_peers_ssl[ip, p];
	else
		return F;

	return peer$hand_over;
	}

function handover_start_processing()
	{
	uninstall_src_net_filter(0.0.0.0/0);
	}

event bro_init()
	{
	# Disable packet processing.
	install_src_net_filter(0.0.0.0/0, 0, 100);
	log "waiting for hand-over - packet processing disabled.";
	}

event remote_connection_error(ip: addr, p: port, reason: string)
	{
	if ( is_remote_event() || ! is_handover_peer(ip, p) )
		return;

	# Seems that the other side in not running.
	log "can't connect for hand-over - starting processing ...";
	handover_start_processing();
	}

event remote_connection_established(ip: addr, p: port)
	{
	if ( is_remote_event() )
		return;

	# If [ip,p] is defined in remote_peers and p != 0, we have connected
	# to the host.
	if ( p != 0/tcp &&
	     ([ip, p] in remote_peers_clear || [ip, p] in remote_peers_ssl) )
		{
		if ( ! is_handover_peer(ip, p) )
			return;

		log fmt("requesting hand-over from %s:%d", ip, p);

		request_remote_events(ip, p, /handover_.*|finished_send_state/);

		# Give the remote side some time to register its handlers.
		schedule 3 secs { handover_request(ip, p) };
		return;
		}

	# If the other side connected to us, we will allow the hand-over
	# if the remote host is defined as a hand-over host in remote_peers.
	if ( is_handover_peer(ip, 0/tcp) )
		{
		log fmt("allowing hand-over from %s:%d", ip, p);
		request_remote_events(ip, p, /handover_.*|finished_send_state/);
		}
	}

event handover_send_state(host: addr, p:port)
	{
	if ( is_remote_event() )
		return;

	# There may be a serialization in progress in which case
	# we will have to try again.
	if ( ! send_state(host, p) )
		{
		log "can't send state; serialization in progress";
		schedule 5 secs { handover_send_state(host, p) };
		}
	}

event handover_request(dst_host: addr, dst_port: port)
	{
	# Make sure the event is for us.
	if ( ! (is_remote_event() && is_it_us(dst_host, dst_port)) )
		return;

	# Send state to other side.
	local e = get_event_source();
	schedule 1 sec { handover_send_state(e$host, e$p) };
	}

event finished_send_state(dst_host: addr, dst_port: port)
	{
	# We will get this event from the remote side.
	# Make sure it's indeed for us.
	if ( ! is_remote_event() )
		 return;

	local e = get_event_source();
	if ( ! is_handover_peer(e$host, e$p) )
		 return;

	log fmt("full state received from %s:%d - starting processing ...",
		e$host, e$p);

	event handover_got_state(e$host, e$p);

	# Start processing.
	handover_start_processing();
	}

event handover_got_state(dst_host: addr, dst_port: port)
	{
	# Make sure the event is for us.
	if ( ! (is_remote_event() && is_it_us(dst_host, dst_port)) )
		return;

	local e = get_event_source();
	log fmt("%s:%d received our state - terminating", e$host, e$p);
	terminate();
	}
